/*
 * nela - a Braille code trainer
 * Copyright (C) 2012  Enrique Matias-Sanchez <quique at unizar.es>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "perkinsWidget.h"
#include <cmath>
#include <cstdlib>
#include <ctime>
#include <QApplication>
#include <QFile>
#include <QKeyEvent>
#include <QMessageBox>
#include <QTextCodec>
#include <QTextStream>
#include <QTimer>
#include <QVBoxLayout>
#include <QDebug>

PerkinsWidget::PerkinsWidget(QWidget *parent) : QWidget(parent)
{
    populateBrailleHash();
    populateLevels();

    setFocusPolicy(Qt::StrongFocus);
    srand(time(0));
    currentLevel = 0;
    currentPosition = 0;
    askedWord = levels[0];
    expectingInput = true;

    boxes.resize(5);
    boxes.fill(QQueue<Word*>());
    boxes[0].enqueue(askedWord);
    currentBox = 0;

    question = new QLabel(levels[0]->printableString());
    question->setAlignment(Qt::AlignCenter);
    question->setFont(QFont("Tiresias PCfont", 96));

    answer = new QLabel();
    answer->setAlignment(Qt::AlignCenter);
    // DejaVu provides an expanded version of the Vera font family aiming
    // for quality and broader Unicode coverage while retaining the
    // original Vera style.  It includes Braille glyphs.
    answer->setFont(QFont("DejaVu Sans", 96));

    picture = new QLabel();
    picture->setAlignment(Qt::AlignCenter);
    picture->setFont(QFont("DejaVu Sans", 384));

    QHBoxLayout *layout = new QHBoxLayout;
    QVBoxLayout *vlayout = new QVBoxLayout;
    vlayout->addWidget(question);
    vlayout->addWidget(answer);
    layout->addLayout(vlayout);
    layout->addWidget(picture);
    setLayout(layout);
}


void PerkinsWidget::keyReleaseEvent(QKeyEvent *keyevent)
{
    if (keyevent->isAutoRepeat() || !expectingInput)
        return;

    int code = keyevent->key();
    QTimer::singleShot(400, this, SLOT(processInput()));

    switch (code) {
    // FIXME The keys should not be hardcoded.

    case Qt::Key_F:
        pressedPoints << "1";
        break;

    case Qt::Key_D:
        pressedPoints << "2";
        break;

    case Qt::Key_S:
        pressedPoints << "3";
        break;

    case Qt::Key_J:
        pressedPoints << "4";
        break;

    case Qt::Key_K:
        pressedPoints << "5";
        break;

    case Qt::Key_L:
        pressedPoints << "6";
        break;

    case Qt::Key_A:
        pressedPoints << "7";
        break;

    case Qt::Key_Ntilde:
        pressedPoints << "8";
        break;

    case Qt::Key_Escape:
        QApplication::quit();
        break;

    }  // switch(code)
}  // keyReleaseEvent(QKeyEvent *keyevent)


QString PerkinsWidget::parsePoints()
{
    pressedPoints.sort();

    // A user could manage to press the same key twice in the allotted time.
    pressedPoints.removeDuplicates();

    QString strPressedPoints = pressedPoints.join("");
    return strPressedPoints;
}

bool PerkinsWidget::checkCharacter(QString strPressedPoints, QChar askedChar)
{
    return strPressedPoints == braillehash[askedChar];
}

void PerkinsWidget::processInput()
{
    // Every keyReleaseEvent generates a QTimer, which launches this slot.
    // The first time this slot is run for a letter, pressedPoints is cleared.
    // The remaining times, there is nothing left to do.
    if (pressedPoints.isEmpty()) {
        return;
    }

    expectingInput = false;
    QString strPressedPoints = parsePoints();
    pressedPoints.clear();

    qDebug() << "Pos:" << currentPosition << "Ch:" << askedWord->character(currentPosition) << "Dots:" << braillehash[askedWord->character(currentPosition)] << "Got: " << strPressedPoints;

    if (checkCharacter(strPressedPoints, askedWord->character(currentPosition))) {
        // The user entered the expected character.
        qDebug() << "The entered char was right!" << askedWord->character(currentPosition);

        QString partialAnswer = answer->text();
        partialAnswer += pointsToUnicode(braillehash[askedWord->character(currentPosition)]);
        answer->setText(partialAnswer);
        ++currentPosition;
        qDebug() << "Let's go for the next character (if any)";
    } else {
        // The entered character was not right.
        // The word passes to the first box.
        qDebug() << "The entered character was not right.";

        boxes[currentBox].dequeue();
        boxes[0].enqueue(askedWord);
        qDebug() << "The word goes back from box" << currentBox << "to box 0  :-(";
        picture->setText(QString("<font color='red'>%1</font>").arg(QChar(10008))); // Heavy ballot x
        // The user will be asked a new word.
        qDebug() << "Let's ask another word\n";
        useNewWord();
    }

    if (currentPosition == askedWord->length()) {
        // The whole word was answered correctly.
        // The word passes to the next box, unless it already was at the last one.
        qDebug() << "The word was correctly finished!";

        picture->setText(QString("<font color='green'>%1</font>").arg(QChar(10004))); // Heavy check mark
        boxes[currentBox].dequeue();
        if (currentBox != boxes.size()-1) {
            qDebug() << "Let's move" << askedWord->string() << "from box" << currentBox << "to box" << currentBox+1;
            boxes[currentBox+1].enqueue(askedWord);
        } else {
            qDebug() << "The word was already at the last level.";
            boxes[currentBox].enqueue(askedWord);
        }

        // The user will be asked a new word.
        qDebug() << "Let's ask another word\n";
        useNewWord();
    }
    expectingInput = true;
}

void PerkinsWidget::useNewWord()
{
    // Show the solution for the previous word
    QString askedBraille("");
    foreach (QChar ch, askedWord->string()) {
        askedBraille.append(pointsToUnicode(braillehash[ch]));
    }
    QString correctAnswer = askedBraille;
    answer->setText(correctAnswer);

    // Wait a couple of seconds so that the student can assimilate his work
    QTimer t;
    t.setSingleShot(true);
    t.start(2000);
    while(t.isActive())
        QApplication::processEvents();

    // Select a new word
    if (boxes[0].isEmpty() && boxes[1].length() <= 1 && boxes[2].length() <= 2
            && !boxes[1].contains(levels.at(currentLevel))
            && !boxes[2].contains(levels.at(currentLevel))) {
        // If box 0 is empty, there is at most 1 word in box 1,
        // 2 words in box 2, and the word in the current level
        // is not one of them, let's go for a new word.
        qDebug() << "Let's go to the next level!" << currentLevel+1;
        ++currentLevel;
        askedWord = levels.at(currentLevel);
        boxes[0].enqueue(askedWord);
        currentBox = 0;
        qDebug() << "A completely new word has been selected.\n";
    } else {
        do {
            currentBox = getRandomBox();
            qDebug() << "Current box:" << currentBox;
        } while (boxes[currentBox].isEmpty());
        askedWord = boxes[currentBox].head();
        qDebug() << "Word selected:" << askedWord->string();
    }

    question->setText(askedWord->printableString());
    answer->clear();
    picture->clear();
    currentPosition = 0;
}

// Braille characters have Unicode codes 10240 - 10495 (0x2800 - 0x28FF).
QChar PerkinsWidget::pointsToUnicode(QString points)
{
    int code = 0;
    foreach(const QString &digit, points) {
        code += pow(2, digit.toInt() - 1);
    }
    code += 10240;
    return QChar(code);
}

int PerkinsWidget::getRandomBox()
{
    double n = rand() / (RAND_MAX+1.0);
    if (n < 0.3478)      // 34.78% chance for box 0 (8/23)
        return 0;
    else if (n < 0.6087) // 26.09% chance for box 1 (6/23)
        return 1;
    else if (n < 0.7826) // 17.39% chance for box 2 (4/23)
        return 2;
    else if (n < 0.9130) // 13.04% chance for box 3 (3/23)
        return 3;
    else                 //  8.70% chance for box 4 (2/23)
        return 4;
}

void PerkinsWidget::populateLevels() {
    #ifdef PKGDATADIR
        QString data_folder = PKGDATADIR;
    #else
        QString data_folder = QApplication::applicationDirPath();
    #endif

    QString shortLocale = QLocale::system().name().left(2);
    if (shortLocale == "C") shortLocale = "en";
    QFile file(data_folder + "/tables/" + shortLocale + ".txt");
    file.setFileName(data_folder + "/levels/" + shortLocale + ".txt");
    if (!file.exists()) {
        file.setFileName(data_folder + "/levels/es.txt");
        QMessageBox::warning(this, tr("Warning"), tr("Unable to find %1 levels, looking for Spanish levels.").arg(QLocale::languageToString(QLocale::system().language())));
    }
    qDebug() << "Levels file:" << file.fileName();

    if (file.open(QIODevice::ReadOnly)) {
        QTextStream in(&file);
        in.setCodec(QTextCodec::codecForName("UTF-8"));
        QString line;
        while (not in.atEnd()) {
            line = in.readLine();
            levels.append(new Word(line));
        }
        file.close();
    } else {
        QMessageBox::critical(this, tr("Critical error"), tr("ERROR: Unable to open file %1").arg(file.fileName()));
        QApplication::quit(); // FIXME It doesn't quit!
    }

    qDebug() << "= LEVELS =";
    foreach(Word* level, levels) {
        qDebug() << level->string();
    }
}

void PerkinsWidget::populateBrailleHash() {
    #ifdef PKGDATADIR
        QString data_folder = PKGDATADIR;
    #else
        QString data_folder = QApplication::applicationDirPath();
    #endif

    QString shortLocale = QLocale::system().name().left(2);
    if (shortLocale == "C") shortLocale = "en";
    QFile file(data_folder + "/tables/" + shortLocale + ".txt");
    if (!file.exists()) {
        file.setFileName(data_folder + "/tables/es.txt");
        QMessageBox::warning(this, tr("Warning"), tr("Unable to find %1 table, looking for Spanish table.").arg(QLocale::languageToString(QLocale::system().language())));
    }
    qDebug() << "Table file:" << file.fileName();

    if (file.open(QIODevice::ReadOnly)) {
        QTextStream in(&file);
        in.setCodec(QTextCodec::codecForName("UTF-8"));
        QString line;
        QChar character;
        QString points;

        while (not in.atEnd()) {
            line = in.readLine();
            QStringList data = line.split(' ');
            character = data[0][0];
            points = data[1];
            braillehash.insert(character, points);
        }
        file.close();
    } else {
        QMessageBox::critical(this, tr("Critical error"), tr("ERROR: Unable to open the table %1").arg(file.fileName()));
        QApplication::quit(); // FIXME It doesn't quit!
    }

    qDebug() << "= TABLE =";

    QHash<QChar, QString>::const_iterator iter = braillehash.constBegin();
    const QHash<QChar, QString>::const_iterator end = braillehash.constEnd();
    for (; iter != end; ++iter) {
        qDebug() << (*iter);
    }
}
